# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
import ssl
import websockets
from typing import Optional, Union
from avassa_client import Session

logger = logging.getLogger(__name__)

logging.getLogger('websockets').setLevel(os.environ.get('WEBSOCKETS_LOGLEVEL', 'WARNING'))


class Topic(object):
    def __init__(self, enum, topic, child_site=None):
        self._enum = enum
        self._topic = topic
        self._child_site = child_site

    @classmethod
    def local(cls, topic: str) -> 'Topic':
        return cls('local', topic)

    @classmethod
    def parent(cls, topic: str) -> 'Topic':
        return cls('parent', topic)

    @classmethod
    def child_site(cls, topic: str, child_site: str) -> 'Topic':
        return cls('child-site', topic, child_site)

    def get_fields(self):
        d = {'location': self._enum,
             'topic': self._topic}
        if self._child_site is not None:
            d['child-site'] = self._child_site
        return d


class Position(object):
    def __init__(self, enum, key=None, value=None):
        self._enum = enum
        self._key = key
        self._value = value

    @classmethod
    def beginning(cls) -> 'Position':
        return cls('beginning')

    @classmethod
    def end(cls) -> 'Position':
        return cls('end')

    @classmethod
    def unread(cls) -> 'Position':
        return cls('unread')

    @classmethod
    def since(cls, duration: str) -> 'Position':
        return cls('since', 'position-since', duration)

    @classmethod
    def seqno(cls, sequence_number: int) -> 'Position':
        return cls('seqno', 'position-sequence-number', sequence_number)

    @classmethod
    def timestamp(cls, timestamp: int) -> 'Position':
        return cls('timestamp', 'position-timestamp', timestamp)

    def get_fields(self):
        if self._key is not None:
            return {self._key: self._value}
        return {'position': self._enum}

class CreateOptions(object):
    def __init__(self, enum, opts=None):
        self._enum = enum
        self._opts = opts

    @classmethod
    def wait(cls) -> 'CreateOptions':
        return cls('wait')

    @classmethod
    def fail(cls) -> 'CreateOptions':
        return cls('fail')

    @classmethod
    def create(cls,
               fmt: str,
               replication_factor: Optional[int] = None,
               persistence: Optional[str] = None,
               local_placement: Optional[bool] = None,
               num_chunks: Optional[int] = None,
               ephemeral: Optional[bool] = None) -> 'CreateOptions':
        create_opts = {'format': fmt}
        if replication_factor is not None:
            create_opts['replication-factor'] = replication_factor
        if persistence is not None:
            create_opts['persistence'] = persistence
        if local_placement is not None:
            create_opts['local-placement'] = local_placement
        if num_chunks is not None:
            create_opts['num-chunks'] = num_chunks
        if ephemeral is not None:
            create_opts['ephemeral'] = ephemeral
        return cls('create', create_opts)

    def get_fields(self):
        d = {'on-no-exists': self._enum}
        if self._enum == 'create':
            d['create-options'] = self._opts
        return d


class ResponseDecodeError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message


class ResponseError(Exception):
    def __init__(self, message: str, errors: list) -> None:
        self.message = message
        self.errors = errors


def decode_result(res):
    try:
        msg = json.loads(res.decode('utf-8'))
        result = msg['result']
    except Exception as e:
        raise ResponseDecodeError("Unable to decode response") from e
    if result == 'error':
        errors = msg.get('errors', [])
        if len(errors) > 1:
            raise ResponseError("Multiple errors", errors)
        elif len(errors) == 1:
            raise ResponseError(errors[0].get('error-message', 'Unknown error'), errors)
        else:
            raise ResponseError("Unknown error", errors)
    elif result != 'ok':
        raise ResponseError(f'Strange result: {result}', None)
    return msg


class Producer(object):
    def __init__(self,
                 session: Session,
                 producer_name: str,
                 topic: Topic,
                 on_no_exists: CreateOptions) -> None:
        self._session = session
        self._producer_name = producer_name
        self._topic = topic
        self._on_no_exists = on_no_exists
        self.log = logger.getChild(__class__.__name__)

    async def _close(self):
        la = self._ws.local_address
        await self._ws.close()
        self.log.debug('Closed producer socket %s', la)

    async def __aenter__(self):
        ssl_context = self._session.get_ssl_context()
        hostname = self._session.get_server_hostname()
        self._ws = await websockets.connect(self._session.get_volga_ws_url(),
                                            extra_headers=self._session.auth_header(),
                                            ssl=ssl_context, server_hostname=hostname)
        try:
            self.log.debug('Opened producer socket %s', self._ws.local_address)
            await self._open_producer()
            return self
        except Exception:
            await self._close()
            raise

    async def __aexit__(self, *args, **kwargs):
        await self._close()

    async def _open_producer(self) -> None:
        cmd = {
            "op": "open-producer",
            "name": self._producer_name,
        }
        cmd.update(self._topic.get_fields())
        cmd.update(self._on_no_exists.get_fields())

        self.log.debug("_open_producer payload: %s", cmd)
        await self._ws.send(json.dumps(cmd, ensure_ascii=False).encode('utf-8'))

        res = await self._ws.recv()
        self.log.debug("_open_producer res: %s", res)
        decode_result(res)

    async def produce(self, payload: Union[str, dict]) -> None:
        cmd = {
            "op": "produce",
            "sync": "sync",
            "payload": payload
        }
        self.log.debug("produce payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))
        res = await self._ws.recv()
        self.log.debug("produce res: %s", res)
        return res.decode('utf-8')

    async def recv(self):
        return await self._ws.recv()

    async def ping(self, data):
        return await self._ws.ping(data)


class InfraProducer(object):
    def __init__(self,
                 session: Session,
                 producer_name: str,
                 infra: str) -> None:
        self._session = session
        self._producer_name = producer_name
        self._infra = infra
        self.log = logger.getChild(__class__.__name__)

    async def _close(self):
        la = self._ws.local_address
        await self._ws.close()
        self.log.debug('Closed infra producer socket %s', la)

    async def __aenter__(self):
        ssl_context = self._session.get_ssl_context()
        hostname = self._session.get_server_hostname()
        self._ws = await websockets.connect(self._session.get_volga_ws_url(),
                                            extra_headers=self._session.auth_header(),
                                            ssl=ssl_context, server_hostname=hostname)
        try:
            self.log.debug('Opened infra producer socket %s', self._ws.local_address)
            await self._open_infra_producer()
            return self
        except Exception:
            await self._close()
            raise

    async def __aexit__(self, *args, **kwargs):
        await self._close()

    async def _open_infra_producer(self) -> None:
        cmd = {
            "op": "open-infra-producer",
            "infra": self._infra
        }

        self.log.debug("_open_infra_producer payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))

        res = await self._ws.recv()
        self.log.debug("_open_infra_producer res: %s", res)
        decode_result(res)

    async def produce(self,
                      direction: str,
                      payload: Union[str, dict],
                      local_site_deliver: Optional[str]=None,
                      destination_sites: Optional[list]=None) -> None:
        cmd = {
            "op": "infra-produce",
            "direction": direction,
            "payload": payload
        }
        if local_site_deliver:
            cmd['local-site-deliver'] = local_site_deliver
        if destination_sites:
            dst = [{'site': x} for x in destination_sites]
            cmd['destination-sites'] = dst

        self.log.debug("produce payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))
        res = await self._ws.recv()
        self.log.debug("produce res: %s", res)
        return res.decode('utf-8')

    async def recv(self):
        return await self._ws.recv()

    async def ping(self, data):
        return await self._ws.ping(data)


class Consumer(object):
    def __init__(self,
                 session: Session,
                 consumer_name: str,
                 mode: str,
                 position: Position,
                 topic: Topic,
                 on_no_exists: CreateOptions,
                 end_marker: Optional[bool] = None) -> None:
        self._session = session
        self._topic = topic
        self._consumer_name = consumer_name
        self._mode = mode
        self._position = position
        self._on_no_exists = on_no_exists
        self._end_marker = end_marker
        self.log = logger.getChild(__class__.__name__)
        self.last_seqno = 0

    async def _close(self):
        la = self._ws.local_address
        await self._ws.close()
        self.log.debug('Closed consumer socket %s', la)

    async def __aenter__(self):
        ssl_context = self._session.get_ssl_context()
        hostname = self._session.get_server_hostname()
        self._ws = await websockets.connect(self._session.get_volga_ws_url(),
                                            extra_headers=self._session.auth_header(),
                                            ssl=ssl_context, server_hostname=hostname)
        try:
            self.log.debug('Opened consumer socket %s', self._ws.local_address)
            await self._open_consumer()
            return self
        except Exception:
            await self._close()
            raise

    async def __aexit__(self, *args, **kwargs):
        await self._close()

    async def _open_consumer(self) -> None:
        cmd = {
            "op": "open-consumer",
            "name": self._consumer_name,
            "mode": self._mode,
        }
        cmd.update(self._topic.get_fields())
        cmd.update(self._position.get_fields())
        cmd.update(self._on_no_exists.get_fields())
        if self._end_marker is not None:
            cmd['end-marker'] = self._end_marker
        self.log.debug("_open_consumer payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))

        res = await self._ws.recv()
        self.log.debug("_open_consumer res: %s", res)
        decode_result(res)

    async def more(self, n):
        cmd = {
            "op": "more",
            "n": n
        }
        self.log.debug("more payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))

    async def recv(self, auto_more=10):
        msg_raw = await self._ws.recv()
        msg = json.loads(msg_raw.decode('utf-8'))
        self.log.debug("recv %s", msg_raw)
        if auto_more and (msg['remain'] == 0):
            await self.more(auto_more)
        seqno = msg.get('seqno')
        if seqno is not None:
            self.last_seqno = seqno
        return msg

    async def ack(self, seqno=None):
        """Acknowledge a message. If no seqno is provided, acknowledge the most
        recently received message."""
        if seqno is None:
            seqno = self.last_seqno
        cmd = {
            "op": "ack",
            "seqno": seqno
        }
        self.log.debug("ack payload: %s", cmd)
        await self._ws.send(json.dumps(cmd, ensure_ascii=False).encode('utf-8'))


class InfraConsumer(Consumer):
    def __init__(self,
                 session: Session,
                 consumer_name: str,
                 mode: str,
                 position: Position,
                 infra: str,
                 end_marker: Optional[bool] = None) -> None:
        self._session = session
        self._infra = infra
        self._consumer_name = consumer_name
        self._mode = mode
        self._position = position
        self._end_marker = end_marker
        self.log = logger.getChild(__class__.__name__)
        self.last_seqno = 0

    async def _close(self):
        la = self._ws.local_address
        await self._ws.close()
        self.log.debug('Closed infra consumer socket %s', la)

    async def __aenter__(self):
        ssl_context = self._session.get_ssl_context()
        hostname = self._session.get_server_hostname()
        self._ws = await websockets.connect(self._session.get_volga_ws_url(),
                                            extra_headers=self._session.auth_header(),
                                            ssl=ssl_context, server_hostname=hostname)
        try:
            self.log.debug('Opened infra consumer socket %s', self._ws.local_address)
            await self._open_infra_consumer()
            return self
        except Exception:
            await self._close()
            raise

    async def __aexit__(self, *args, **kwargs):
        await self._close()

    async def _open_infra_consumer(self) -> None:
        cmd = {
            "op": "open-infra-consumer",
            "name": self._consumer_name,
            "mode": self._mode,
            "infra": self._infra
        }
        cmd.update(self._position.get_fields())
        if self._end_marker is not None:
            cmd['end-marker'] = self._end_marker

        self.log.debug("_open_infra_consumer payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))

        res = await self._ws.recv()
        self.log.debug("_open_infra_consumer res: %s", res)
        decode_result(res)

    async def more(self, n):
        cmd = {
            "op": "more",
            "n": n
        }
        self.log.debug("more payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))

    async def recv(self, auto_more=10):
        msg_raw = await self._ws.recv()
        msg = json.loads(msg_raw.decode('utf-8'))
        self.log.debug("recv %s", msg_raw)
        if auto_more and (msg['remain'] == 0):
            await self.more(auto_more)
        seqno = msg.get('seqno')
        if seqno is not None:
            self.last_seqno = seqno
        return msg

    async def ack(self, seqno=None):
        """Acknowledge a message. If no seqno is provided, acknowledge the most
        recently received message."""
        if seqno is None:
            seqno = self.last_seqno
        cmd = {
            "op": "ack",
            "seqno": seqno
        }
        self.log.debug("ack payload: %s", cmd)
        await self._ws.send(json.dumps(cmd, ensure_ascii=False).encode('utf-8'))


class QueryConsumer(object):
    def __init__(self,
                 session: Session,
                 query: dict):
        self._query = query
        self._session = session
        self.log = logger.getChild(__class__.__name__)

    async def _close(self):
        la = self._ws.local_address
        await self._ws.close()
        self.log.debug('Closed query consumer socket %s', la)

    async def __aenter__(self):
        ssl_context = self._session.get_ssl_context()
        hostname = self._session.get_server_hostname()
        self._ws = await websockets.connect(self._session.get_volga_ws_url(),
                                            extra_headers=self._session.auth_header(),
                                            ssl=ssl_context, server_hostname=hostname)
        try:
            self.log.debug('Opened query consumer socket %s', self._ws.local_address)
            await self._open_query_consumer()
            return self
        except Exception:
            await self._close()
            raise

    async def __aexit__(self, *args, **kwargs):
        await self._close()

    async def _open_query_consumer(self) -> None:
        cmd = {
            "op": "query-topics"
        }
        cmd.update(self._query)

        self.log.debug("_open_query_consumer payload: %s", cmd)
        await self._ws.send(json.dumps(cmd).encode('utf-8'))

    async def recv(self):
        try:
            data = await self._ws.recv()
            self.log.debug("data %s", data)
            return data
        except websockets.exceptions.ConnectionClosedOK:
            return None


async def delete_topic(session: Session, topic: str):
    ssl_context = self._session.get_ssl_context()
    hostname = self._session.get_server_hostname()
    async with websockets.connect(session.get_volga_ws_url(),
                                  extra_headers=session.auth_header(),
                                  ssl=ssl_context, server_hostname=hostname) as ws:
        cmd = {'op': 'delete-topic',
               'topic': topic}
        logger.debug('delete_topic payload: %s', cmd)
        await ws.send(json.dumps(cmd).encode('utf-8'))
        res = await ws.recv()
        logger.debug('delete_topic res: %s', res)
        return res.decode('utf-8')



# SPDX-License-Identifier: Apache-2.0

import base64
import json
import logging
import os
import ssl
import sys
import urllib.request
from typing import Dict, Optional
from urllib.parse import urlparse

logger = logging.getLogger(__name__)
# logging.getLogger().setLevel(logging.DEBUG)


class Error(Exception):
    pass


class LoginError(Error):
    def __init__(self, message: str) -> None:
        self.message = message


class GetError(Error):
    def __init__(self, message: str) -> None:
        self.message = message


class PostError(Error):
    def __init__(self, message: str) -> None:
        self.message = message


class ApiSSLContext(ssl.SSLContext):
    def __new__(cls, hostname, **kwargs):
        return super(ApiSSLContext, cls).__new__(cls, ssl.PROTOCOL_TLS_CLIENT, **kwargs)

    def __init__(self, hostname, **kwargs):
        super(ApiSSLContext, self).__init__()
        self._server_hostname = hostname
        self.options |= ssl.OP_NO_SSLv2
        self.options |= ssl.OP_NO_SSLv3
        self.verify_mode = ssl.CERT_REQUIRED
        self.check_hostname = True

    def wrap_socket(self, *args, **kwargs):
        if self._server_hostname is not None:
            kwargs['server_hostname'] = self._server_hostname
        return super().wrap_socket(*args, **kwargs)


class Session(object):
    def __init__(self, host: str,
                 token: str,
                 ssl_context: Optional[ssl.SSLContext] = None,
                 user_agent: Optional[str] = None,
                 server_hostname: Optional[str] = None) -> None:
        self._server_hostname = server_hostname
        self._base_url = urlparse(host)
        self._volga_ws_url = f"wss://{self._base_url.netloc}/v1/ws/volga"
        self._token = token
        if ssl_context:
            if server_hostname is not None:
                raise ValueError("server_hostname is not applicable when"
                                 " using custom ssl_context")
            self._ssl_context = ssl_context
        else:
            self._ssl_context = create_ssl_context(server_hostname)
        self._user_agent = user_agent

    def auth_header(self) -> Dict[str, str]:
        header = {"Authorization": "Bearer {}".format(self._token)}
        if self._user_agent:
            header['User-Agent'] = self._user_agent
        return header

    def get_token(self) -> str:
        return self._token

    def get_base_url(self) -> str:
        return self._base_url

    def get_user_agent(self) -> str:
        return self._user_agent

    def get_volga_ws_url(self) -> str:
        return self._volga_ws_url

    def get_ssl_context(self):
        return self._ssl_context

    def get_server_hostname(self) -> str:
        return self._server_hostname


def create_ssl_context(server_hostname=None):
    c = ApiSSLContext(server_hostname)
    c.verify_mode = ssl.CERT_REQUIRED
    c.check_hostname = True
    if 'API_CA_CERT' in os.environ:
        c.load_verify_locations(cadata=os.environ['API_CA_CERT'])
    else:
        c.load_default_certs(ssl.Purpose.SERVER_AUTH)
    if hasattr(c, 'keylog_filename'):
        keylogfile = os.environ.get('SSLKEYLOGFILE')
        if keylogfile and not sys.flags.ignore_environment:
            c.keylog_filename = keylogfile
    return c


def get_request(session: Session, url, user_agent=None, ssl_context=None, headers=None):
    return _send_request('GET', session, url, None, user_agent, ssl_context, headers)


def post_request(session: Session, url, payload, user_agent=None, ssl_context=None, headers=None):
    return _send_request('POST', session, url, payload, user_agent, ssl_context, headers)


def put_request(session: Session, url, payload, user_agent=None, ssl_context=None, headers=None):
    return _send_request('PUT', session, url, payload, user_agent, ssl_context, headers)


def _send_request(method, session: Session, url, payload, user_agent, ssl_context, headers):
    if headers is None:
        headers = {}
    if 'Accept' not in headers:
        headers['Accept'] = 'application/json'
    if session:
        if user_agent is None:
            user_agent = session.get_user_agent()
        if ssl_context is None:
            ssl_context = session.get_ssl_context()
        headers['Authorization'] = 'Bearer ' + session.get_token()
        url = urllib.parse.urljoin(session.get_base_url().geturl(), url)
    if user_agent:
        headers['User-Agent'] = user_agent
    if payload is None:
        data = None
        logger.debug('%s request url %s', method, url)
    else:
        headers['Content-Type'] = 'application/json'
        data = json.dumps(payload).encode()
        logger.debug('%s request url %s, payload %s', method, url, data)
    req = urllib.request.Request(url, method=method, headers=headers, data=data)
    with urllib.request.urlopen(req, context=ssl_context) as response:
        return (response.getcode(),
                response.msg,
                response.getheaders(),
                response.read().decode('utf-8'))


def approle_login(host: str,
                  role_id: str | None,
                  secret_id: str,
                  user_agent: Optional[str] = None,
                  server_hostname: Optional[str] = None):
    try:
        payload = {'secret-id': secret_id}
        if role_id is not None:
            payload['role-id'] = role_id
        base_url = urlparse(host)
        url = base_url.geturl() + '/v1/approle-login'
        sslc = create_ssl_context(server_hostname)
        (_code, msg, _headers, body) = post_request(
            None, url, payload, user_agent, sslc)
        del payload['secret-id']
        if msg != 'OK':
            raise LoginError('Approle login failed: ' + msg)
        return Session(host, json.loads(body)['token'], sslc,
                       user_agent=user_agent)
    except urllib.error.HTTPError as e:
        raise LoginError('Exception during approle_login: ' +
                         get_error(e)) from e
    except urllib.error.URLError as e:
        raise LoginError('Exception during approle_login: ' +
                         get_error(e)) from e
    except Exception as e:
        raise LoginError('Exception during approle_login') from e


def get_error(e):
    try:
        body = json.loads(e.__dict__['file'].read().decode('utf-8'))
        return body['errors'][0]['error-message']
    except Exception:
        return "unknown"


def kubernetes_login(host: str,
                     tenant: str,
                     service: str,
                     role: str,
                     jwt: str,
                     user_agent: Optional[str] = None,
                     check_hostname: bool = True,
                     server_hostname: Optional[str] = None):
    try:
        payload = {"tenant": tenant,
                   "service": service,
                   "role": role,
                   "jwt": jwt}
        base_url = urlparse(host)
        url = base_url.geturl() + '/v1/kubernetes-login'
        sslc = create_ssl_context(server_hostname)
        sslc.check_hostname = check_hostname
        (_code, msg, _headers, body) = post_request(
            None, url, payload, user_agent, sslc)
        if msg != 'OK':
            raise LoginError("Kubernetes login failed: " + msg)
        return Session(host, json.loads(body)['token'], sslc,
                       user_agent=user_agent)
    except urllib.error.HTTPError as e:
        raise LoginError('Exception during kubernetes_login: ' +
                         get_error(e)) from e
    except urllib.error.URLError as e:
        raise LoginError('Exception during kubernetes_login ' +
                         get_error(e)) from e
    except Exception as e:
        raise LoginError('Exception during kubernetes_login') from e


def login(host: str,
          username: str,
          password: str,
          tenant: Optional[str] = None,
          user_agent: Optional[str] = None,
          server_hostname: Optional[str] = None) -> Session:
    try:
        payload = {'username': username,
                   'password': password}
        if tenant:
            payload['tenant'] = tenant

        base_url = urlparse(host)
        url = base_url.geturl() + '/v1/login'
        sslc = create_ssl_context(server_hostname)
        (_code, msg, _headers, body) = post_request(
            None, url, payload, user_agent, sslc)
        del payload['password']
        if msg != "OK":
            raise LoginError('Login failed: ' + msg)
        return Session(host, json.loads(body)['token'], sslc,
                       user_agent=user_agent)
    except urllib.error.HTTPError as e:
        raise LoginError('Exception during login: ' +
                         get_error(e)) from e
    except urllib.error.URLError as e:
        raise LoginError('Exception during login ' +
                         get_error(e)) from e
    except Exception as e:
        raise LoginError('Exception during login') from e


# Example Strongbox integrations

def get_secret(session: Session,
               vault_name: str,
               secret_name: str) -> Dict[str, str]:
    url = "{}/v1/state/strongbox/vaults/{}/secrets/{}".format(
        session.get_base_url().geturl(), vault_name, secret_name)
    (_code, msg, _headers, body) = get_request(session, url)
    if msg != 'OK':
        raise GetError('Get secret failed: ' + msg)
    return json.loads(body)['dict']


def encrypt(session: Session, key_name: str, plain_text: str) -> str:
    url = "{}/v1/state/strongbox/transit-keys/{}/encrypt".format(
        session.get_base_url().geturl(), key_name)
    payload = {
        "plaintext": str(base64.b64encode(plain_text.encode('utf-8')),
                         'utf-8')
    }
    (_code, msg, _headers, body) = post_request(session, url, payload)
    del payload['plaintext']
    if msg != "OK":
        raise PostError("Encrypt failed: " + msg)
    return json.loads(body)["ciphertext"]


def decrypt(session: Session, key_name: str, cipher_text: str) -> str:
    url = "{}/v1/state/strongbox/transit-keys/{}/decrypt".format(
        session.get_base_url().geturl(), key_name)
    payload = {
        "ciphertext": cipher_text,
    }
    (_code, msg, _headers, body) = post_request(session, url, payload)
    if msg != "OK":
        raise PostError("Decrypt failed: " + msg)
    return str(base64.b64decode(json.loads(body)["plaintext"]),
               'utf-8')
